//
//								MOVEOUT
// Visualization program for seismic CMP gathers stored in SEG-Y format 
// http://www.seg.org/documents/10161/77915/seg_y_rev1.pdf
// applies NMO corrections based on a power-law velocity function
//
//
// Copyright 2011-2015 by Yuri Ganshin, CMI, University of Wyoming, Laramie.
// http://www.uwyo.edu/cmi/
//
// This program is free sofware released using the Fast Light Tool Kit (FLTK).
// For conditions of distribution and use, see the license agreement in 
//     http://www.fltk.org/COPYING.php
// As a general rule, "free software" means that the user have the freedom
// to run, copy, distribute, study, modify and improve the software.
//
// Please report all bugs and problems on the following page:
// http://www.uwyo.edu/cmi/
//     
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <cfloat>  // declares FLT_MAX
//
#include <FL/Fl.H>
#include <FL/Fl_Window.H>
#include <FL/Fl_Double_Window.H>
#include <FL/fl_draw.H>	
#include <FL/Fl_Menu_Bar.H>
#include <FL/Fl_Native_File_Chooser.H>
#include <FL/Fl_Int_Input.H>
// #include <FL/Fl_Slider.H>
#include <FL/Fl_Spinner.H>
#include <FL/Fl_Float_Input.H>
#include <FL/Fl_Input_Choice.H>
#include <FL/fl_show_colormap.H>
#include <FL/Fl_Output.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Box.H>
#include <FL/fl_message.H>
#include <FL/fl_ask.H> // for fl_alert() and fl_message()
#include <math.h>

using namespace std;

#define TIME_BAR	60
#define HDR_BAR		24
#define RIM			 8	// RIM == border for canvas
#define MAX_VAL		100	// Maximum number of trace headers to plot
//
#ifndef FLT_MAX
#define FLT_MAX 1e+37
#endif
//
#ifndef M_PI
#define M_PI 3.14159265359
#endif
//
////////////////////////////////////////////////////////////////
/****               List of GLOBAL variables               ****/
FILE *in_fp = NULL;
char inpname[200];
int nsamp, dsamp, Ntr, convert_mode=1;
short measure;
int tr_sta,tr_fin,samp_sta,samp_fin,hdr_step=5;
int tr_vis=-1,samp_vis,wiggle_trace=0;
int numpix=3;
int hdr_val[MAX_VAL];
//
int canvas_w = 760, canvas_h =960;
//
int word_no=5;				// CDP for header plot
int nmo_no=1;				// NMO correction method, 1 - hyperbolic
unsigned int tr_size, samp_step=5;
float Norm, gain, elev_scal, coor_scal, *seis;
float maxSF=17.0, nu = 0.0;
char *hdrs,c_word[4],nmo_word[3],cgbuf[80];
char text_hdr[3201];
double v0=2100.0, alfa=1350.0, beta=0.60;

////////////////////////////////////////////////////////////////
/****            List of Function Prototypes               ****/

void read_file_hdr(const char *name);
void read_seis_dat();
void convert2(char* input);
void convert4(char* input);
void convert_hdrs(char* array);
float ibm2ieee(const float ibm);
float efloat(float value);
unsigned char to_ascii ( unsigned char c );
float conv3(float *dat,int NS1,int NS2,int index);
void pal_interp(float [], int, float, short []);

////////////////////////////////////////////////////////////////
/****         Global pointers to the GUI objects           ****/
////////////////////////////////////////////////////////////////
Fl_Double_Window *seis_window, *proc_window;
Fl_Window *opts_window, *hdr_window;
Fl_Int_Input	*tr_sta_field, *tr_fin_field;
Fl_Int_Input	*t_sta_field,  *t_fin_field;
Fl_Int_Input	*t_pick_field, *v_pick_field;
Fl_Float_Input	*v0_field, *alfa_field, *beta_field; 
Fl_Float_Input *gain_field, *maxs_field, *nu_field;
Fl_Button *apply, *next_tr, *add_wiggle, *show_hdr, *end_this;
Fl_Button *apply_nmo;

////////////////////////////////////////////////////////////////
/****        List of Widgets and Drawing functions         ****/
////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
/********             File-choosing MENU               ********/
////////////////////////////////////////////////////////////////
class File_menu : public Fl_Window {
	Fl_Native_File_Chooser *fc;
	
	void read_data(const char *filename)	// Process the chosen file
	{
		strcpy(inpname,filename);
		tr_vis=1;			// will wait while(tr_vis<0)
	}
	
	// Handle an 'Open' request from the menu
	static void open_cb(Fl_Widget *w, void *v) {
		File_menu *choice = (File_menu*)v;
		choice->fc->title("Open");
		choice->fc->type(Fl_Native_File_Chooser::BROWSE_FILE);
		switch ( choice->fc->show() ) {
			case -1: break;	// Error
			case  1: break; // Cancel
			default:		// choice->fc->show() returns 0
				choice->read_data(choice->fc->filename());
				// choice->deactivate();				
				break;
		}
	}

	static void quit_cb(Fl_Widget *w, void *v) {
		exit(0);
	}
	
public:
	File_menu() : Fl_Window(500,300,320,160,"Choose the file") {
		Fl_Menu_Bar *menu = new Fl_Menu_Bar(0,0,300,25);
		menu->add("&File/&Open",  FL_COMMAND+'o', open_cb, (void*)this);
		menu->add("&File/&Quit",  FL_COMMAND+'q', quit_cb);
		
		// Describe the chooser
		Fl_Box *box = new Fl_Box(20,45,w()-40,h()-65);
		box->color(FL_GRAY-1); 
		box->box(FL_FRAME_BOX);
		box->align(FL_ALIGN_CENTER|FL_ALIGN_INSIDE|FL_ALIGN_WRAP);
		box->label("Choose seismic binary SEG-Y file\n\
with 4-byte floating point data format\n\
[ IBM, IEEE, or native byte order ]");
		box->labelsize(16);
		// Initialize the file chooser
		fc = new Fl_Native_File_Chooser();
		fc->filter("SEG-Y Binary\t*.sgy\n");			 
		end();
	}
};
//////////////////////////////////////////////////////////////////////
/**************        Drawing Seismic Data          ****************/
//   Double-buffered window eliminates flickering during updates
//////////////////////////////////////////////////////////////////////
class Draw_seis : public Fl_Double_Window 
{
	void draw();
public:
  	Draw_seis(int X,int Y,int W,int H) : Fl_Double_Window(X,Y,W,H) {}
	
} *seismic; 

void Draw_seis::draw()  // Function draw() within class Draw_seis
{      /**    don't declare reserved 'x' and 'y' variables !!!   **/
	Fl_Double_Window::draw();
	unsigned i, j;
	int xx, xx1=0, yy, x_dim, y_dim, *hdr4;
	int val, tr1, tr2, sta_pix;
	float yi, a1, a2;
	short *hdr2;
	double dx,dtr;
	
	if(samp_vis<50) samp_step=5;		// Step for Time Labels
	else if(samp_vis<500) samp_step=25;
	else if(samp_vis<1000) samp_step=50;	
	else if(samp_vis<2000) samp_step=100;
	else if(samp_vis<5000) samp_step=250;
	else if(samp_vis<10000) samp_step=500;
	else samp_step=1000;

	hdr2=(short*)hdrs;	// To get trace ID (0=Unknown,1=Seismic,2=Dead)
   hdr4=(int*)hdrs;
	for(tr1=0,j=0;tr1<tr_vis;tr1+=hdr_step,j++) 
	{
		val = hdr4[tr1*60+word_no];
		if(word_no>=10 && word_no<=14) val=(int)((float)val*elev_scal);
		if(word_no>=18 && word_no<=46) val=(int)((float)val*coor_scal);
		if(j<MAX_VAL) hdr_val[j]=val;
	}			

	/**    Estimate X-shft from the number of numerals	  **/
	tr1=tr_vis/2;      // take middle trace as estimate
	val = hdr4[tr1*60+word_no];  // 60 = number_of_4-byte_words
	if(word_no>=10 && word_no<=14) val=(int)((float)val*elev_scal);
	if(word_no>=18 && word_no<=46) val=(int)((float)val*coor_scal);
	if(val<10) numpix=3;
	else if(val<100) numpix=7;
	else if(val<1000) numpix=11;
	else if(val<10000) numpix=15;		
	else if(val<100000) numpix=19;
	else if(val<1000000) numpix=23;
	else numpix=27;			//  numpix == X-shift in pixels
		
	x_dim = w() - TIME_BAR;  y_dim = h() - HDR_BAR;
	//cout << "Xdim = " << x_dim << ", Y_dim = " << y_dim << endl;	
	/////////////////////////////////////////////////////////////////

	fl_rectf(0,0,w(),h(),  0,  0,  0);		// Black bounding rectangle
	fl_rectf(1,1,w()-2,h()-2,230,230,230);	// Gray bounding rectangle
	///////    Finally, clean the WORK AREA with WHITE	
	fl_rectf(TIME_BAR,HDR_BAR,x_dim-1,y_dim-1,255,255,255);
	
	fl_color(FL_BLACK);	fl_font(0,14);	
	
	dtr = (double)x_dim/tr_vis;  //  decimal  pixels/trace
	xx1 = sta_pix = TIME_BAR + (int)(dtr/2.0) + 1;
	
	///////////////////////////////////////////////////////////////////////
	// Variable Area Display utilizing cubic convolution for resampling	 //
	//  and linear interpolataion for filling inter-trace space			 //
	///////////////////////////////////////////////////////////////////////	
	for(tr1=0;tr1<tr_vis;tr1++)
	{										
		i = tr1*x_dim/tr_vis;
		xx = i + sta_pix;
		tr2 = tr1+1;
		dx = fabs((double)i - dtr*tr1)/dtr;	//   0.0 <= dx <= 1.0
		for(j=0;j<y_dim;j++)  // possible: j+=3 for fl_yxline() 
		{
			yy = j+HDR_BAR;	// yy = 24 (j==0)
		// Resample data for visualization using Cubic Convolution Method
			if(hdr2[tr1*120+14]==2) a1=0.0;		
			else a1=conv3(seis+tr1*nsamp+samp_sta,samp_vis,y_dim,j);
			if(hdr2[tr2*120+14]==2) a2=0.0;			
			else a2=conv3(seis+tr2*nsamp+samp_sta,samp_vis,y_dim,j);
			yi = a1+(a2-a1)*dx; // Inter-trace linear interpol: a1<=yi<=a2
			val = (int)((yi*gain*15)/Norm);
			if(val>0)fl_xyline(xx, yy, xx+val); // horiz line for VA display
			if(wiggle_trace)					// optional wiggle trace overlay
				if(j>0) fl_line(xx+val,yy,xx1,yy-1);
			xx1 = xx+val;						
		}								
	}		

	for(tr1=0,j=0;tr1<tr_vis;tr1+=hdr_step,j++)	////  Draw Trace Headers 
	{
		i = tr1*x_dim/tr_vis;
		xx = i + sta_pix;
		fl_yxline(xx, HDR_BAR, HDR_BAR-6);		 
		if(j<MAX_VAL) sprintf(cgbuf,"%d",hdr_val[j]); 
		else sprintf(cgbuf,"%d",99999);
		fl_draw(cgbuf,xx-numpix,HDR_BAR-10);	
	}											////  Draw Trace Headers

   /////////////////////////////////////////////////  Draw Time labels
	sprintf(cgbuf,"Time ");  fl_draw(cgbuf,10,HDR_BAR+8);
	sprintf(cgbuf,"[ ms ]");  fl_draw(cgbuf,10,HDR_BAR+26);	
	for(i=0;i<samp_vis;i+=samp_step)
	{
		yy = y_dim*i/samp_vis + HDR_BAR;
		if(i>0) 
		{
			val=(samp_sta+i)*dsamp/1000;
			sprintf(cgbuf,"%5d",val); fl_draw(cgbuf,10,yy+4); // text string
			fl_xyline(TIME_BAR-10,yy,TIME_BAR+x_dim);  // horiz line			
		}
	}

}

/////////////////////////////////////////////////////////////////////
/**************       Drawing Processed Data        ****************/
/////////////////////////////////////////////////////////////////////
class ProcCanvas : public Fl_Widget 
{
public:
	ProcCanvas(int X,int Y,int W,int H) : Fl_Widget(X,Y,W,H) {}

	void draw()
	{
		unsigned i, j, pos;
		int xx, xx1=0, yy, x_dim, y_dim, *hdr4;
		int val, t, tr, tr1, tr2, sta_pix;
		short *hdr2;
		float yi, a1, a2, sir, stretch, maxs, offs;
		float amp, amp1, amp2;
		float *proc,*count,*outbuf;
		double *Vq;
		double dx,dtr,time,vel,t02,v2,v4,Vc,val1,val2,val4,off2,off4;
	
		x_dim = w() - TIME_BAR;  y_dim = h() - HDR_BAR;

		fl_rectf(x()-1,y()-1,w()+2,h()+2,  0,  0,  0);	// Black bounding rectangle
		fl_rectf(x(),  y(),  w(),  h(),  230,230,230);	// Gray bounding rectangle
		///////    Finally, clean the WORK AREA with WHITE	
		fl_rectf(x()+TIME_BAR,y()+HDR_BAR,x_dim,y_dim,255,255,255);
		fl_color(FL_BLACK);	fl_font(0,14);	 // Helvetica regular, 14 pixel
		dtr = (double)x_dim/tr_vis;  //  decimal  pixels/trace
		xx1 = sta_pix = TIME_BAR + (int)(dtr/2.0) + 1;
		
		hdr2=(short*)hdrs;	// To get trace ID (0=Unknown,1=Seismic,2=Dead)
		///////////////////////////////////////////////////////////////////////
		////    Processing    Processing    Processing    Processing	   ////
		///////////////////////////////////////////////////////////////////////
					
		count = new (nothrow) float[nsamp];
		outbuf = new (nothrow) float[nsamp];
		Vq = new (nothrow) double[nsamp];	// Quartic Velocity correction				
		proc = new (nothrow) float[(tr_vis+1) * nsamp];
		if(proc == 0)
		{ 
			fl_message("ERROR: NO memory to process data!\n\
						  Close the window to exit...");	   exit(1);
		}
	// Do not damage original seis[] data, but use temporary array 
		memcpy(proc, seis, (tr_vis+1) * nsamp * sizeof(float));
	
/**  In a constant velocity medium, stretch factor S is commonly estimated
S = sqrt(1 + X*X/(T*T*V*V)); [Vermeer,1990,Seismic wavefield sampling:SEG]
where S is in decimal fractions, X - offset, V = Vrms, T = t0          **/
		maxs = maxSF+100.0;  maxs /= 100.0;	// translate percent to fractions
		val2 = (double)maxs*maxs - 1.0;  maxs = (float)val2;
		
		sir = 1000000.0/dsamp;					// Time samples per second
		
		if(nmo_no==3)
		{	
			val4 = 0.0;
			for(i=0;i<nsamp;i++)
			{
				time = (double)i/sir;		// time in 'sec'
				val1 = pow(time,beta);			
				val2 = pow(time*time,beta);
				// Lumley's Vint formula from the SEP Report #75, 1997
				Vc = sqrt(v0*v0+v0*2*alfa*(beta+1)*val1+(beta*2+1)*alfa*alfa*val2);
				Vc /= sir;						// units/sample
				val4 += Vc*Vc*Vc*Vc;
				Vq[i] = val4/(i+1);  //  Quartic Velocity estimation taken from
				//	 Biondi, et al., "Nonstretch normal moveout ...", Geophys., 2014
			}			
		}
		
		hdr4=(int*)hdrs;
	   for(tr=0;tr<tr_vis;tr++)				// Calculate NMO
		{
			pos = tr*nsamp;
			offs = (float)abs(hdr4[tr*60+9]);  // 60 = number_of_4-byte_words
			off2 = (double)offs*offs;
			off4 = off2*off2;

			for(i=0;i<nsamp;i++) { outbuf[i] = count[i] = 0.0; }
			
			for(i=0;i<nsamp;i++)
			{
				time = (double)i/sir;			// time in 'sec'
				vel = (v0 + alfa*pow(time,beta))/sir;	// units/sample

				t02 = (double)i*i;
				v2 = vel*vel;  v4=v2*v2;
				val2 = t02 + off2/v2;		// Hyperbolic traveltime
				val4 = 2.0/(t02*v4);
						
				time = val2;
				if(nmo_no==2) time -= nu*off4*val4/100.0;
				if(nmo_no==3) time += (v4-Vq[i])*off4/(4.0*t02*v4*v4);
				
				if(time<0.0) continue;
				time = sqrt(time);		// samples

				t = (int)time;	
				if( t<= 0 ) continue; 
				if( t>=(nsamp-2) ) continue;
				
				stretch = off2/(t02*v2);				
				if(stretch>maxs) continue;		// limit NMO stretch	
				
				amp1 = proc[pos+t]; amp2 = proc[pos+t+1];
				amp = amp1 - (time - (float)t)*(amp1 - amp2);
				outbuf[i] += amp;  count[i] += 1.0;
			}
			
			for(i=0;i<nsamp;i++) 				// Apply NMO
			{ 
				amp1=sqrt(count[i]); 
				if(amp1>0.0) amp = outbuf[i]/amp1; else amp = 0.0;
				proc[pos+i] = amp;
			}
		}

		///////////////////////////////////////////////////////////////////////
		// Variable Area Display utilizing cubic convolution for resampling	 //
		// and linear interpolataion for filling inter-trace space			 //
		///////////////////////////////////////////////////////////////////////
		for(tr1=0;tr1<tr_vis;tr1++)
		{		
			i = tr1*x_dim/tr_vis;
			xx = x() + i + sta_pix;
			dx = fabs((double)i - dtr*tr1)/dtr;	//   0.0 <= dx <= 1.0			
			tr2 = tr1+1;
 
			for(j=0;j<y_dim;j++)  // possible: j+=3 for fl_yxline() 
			{
				yy = y()+j+HDR_BAR;	// yy = 24 (j==0)
				if(hdr2[tr1*120+14]==2) a1=0.0;				
				else a1=conv3(proc+tr1*nsamp+samp_sta,samp_vis,y_dim,j);
				if(hdr2[tr2*120+14]==2) a2=0.0;				
				else a2=conv3(proc+tr2*nsamp+samp_sta,samp_vis,y_dim,j);
				
				yi = a1+(a2-a1)*dx; // Inter-trace linear interpol: a1<=yi<=a2
				val = (int)((yi*gain*15)/Norm);
				if(val>0)fl_xyline(xx, yy, xx+val); // horiz line for VA display
				if(wiggle_trace)					// optional wiggle trace overlay
					if(j>0) fl_line(xx+val,yy,xx1,yy-1);
				xx1 = xx+val;						
			}								
		}		
		
		for(tr1=0,j=0;tr1<tr_vis;tr1+=hdr_step,j++)	////  Draw Trace Headers 
		{
			i = tr1*x_dim/tr_vis;
			xx = x() + i + sta_pix;
			fl_yxline(xx, y()+HDR_BAR, y()+HDR_BAR-6);		 
			if(j<MAX_VAL) sprintf(cgbuf,"%d",hdr_val[j]); 
			else sprintf(cgbuf,"%d",99999);
			fl_draw(cgbuf,xx-numpix,y()+HDR_BAR-10);	
		}			
		
		xx = x()+10;  yy = y()+HDR_BAR;				////  Draw Time labels
		sprintf(cgbuf,"Time ");  fl_draw(cgbuf,xx,yy+8);
		sprintf(cgbuf,"[ ms ]"); fl_draw(cgbuf,xx,yy+26);
				
		xx = x()+TIME_BAR;
		for(i=0;i<samp_vis;i+=samp_step)
		{
			yy = y_dim*i/samp_vis + y() + HDR_BAR;
			if(i>0) 
			{
				val=(samp_sta+i)*dsamp/1000;
				sprintf(cgbuf,"%5d",val); fl_draw(cgbuf,x()+10,yy+4);
				fl_xyline(xx-10, yy, xx+x_dim);	  // horiz line		
			}
		}
		
		delete[] proc; delete[] count; delete[] outbuf; delete[] Vq; 

	}	// void draw()
	
} *proc_draw;

////////////////////////////////////////////////////////////////
/************    List of Call Back functions    ***************/
////////////////////////////////////////////////////////////////

void input_tr_sta(Fl_Widget *, void *) {
	tr_sta = atoi(tr_sta_field->value());
}

void input_tr_fin(Fl_Widget *, void *) {
	tr_fin = atoi(tr_fin_field->value());
}

void input_t_sta(Fl_Widget *, void *) {
	samp_sta = atoi(t_sta_field->value())*1000/dsamp;
}

void input_t_fin(Fl_Widget *, void *) {
	samp_fin = atoi(t_fin_field->value())*1000/dsamp;
}

void input_gain(Fl_Widget *, void *) {
	gain = atof(gain_field->value());
}

void input_maxs(Fl_Widget *, void *) {
	maxSF = atof(maxs_field->value());
}

void input_nu(Fl_Widget *, void *) {
	nu = atof(nu_field->value());
}

void spinner_cb(Fl_Widget *w, void *) {
   Fl_Spinner *s = (Fl_Spinner *)w;	
   hdr_step = (int)s->value();
	// printf("Spinner Value = %d\n",hdr_step);
}

void input_choice_cb(Fl_Widget*,void*data) {
	Fl_Input_Choice *in_hdr=(Fl_Input_Choice *)data;
	
	strncpy(c_word,in_hdr->value(),3);
	sscanf(c_word,"%d", &word_no);
}

void nmo_choice_cb(Fl_Widget*,void*data) {
	Fl_Input_Choice *in_nmo=(Fl_Input_Choice *)data;
	
	strncpy(nmo_word,in_nmo->value(),2);
	sscanf(nmo_word,"%d", &nmo_no);
}

void show_hdr_cb(Fl_Widget *, void *) 
{
   // fl_message(text_hdr); -> accomdates only 12 rows out of 40
	hdr_window = new Fl_Window(676,676,"File's Textual header");

	Fl_Box* txt = new Fl_Box(FL_FRAME_BOX,8,8,660,660,text_hdr);
	txt->color(FL_WHITE);	
	txt->labelfont(0);
	txt->labelsize(14);
	
	hdr_window->resizable(txt); 
	hdr_window->end();
	hdr_window->show();
}

void input_v0(Fl_Widget *, void *) {
	v0 = atof(v0_field->value());
}

void input_alfa(Fl_Widget *, void *) {
	alfa = atof(alfa_field->value());
}

void input_beta(Fl_Widget *, void *) {
	beta = atof(beta_field->value());
}

void exit_b(Fl_Widget *, void *) {
	exit(0);
}

// Redraw display with updated data
void doRedraw(Fl_Widget *, void*)
{
	if(tr_fin>Ntr) tr_fin=Ntr;	
	if(tr_sta>=tr_fin) tr_sta=1; 
	tr_vis = tr_fin-tr_sta+1; // if trace count starts from '1'

	if(samp_fin>nsamp) samp_fin=nsamp;
	if(samp_sta>=samp_fin) samp_sta=0;
	samp_vis = samp_fin-samp_sta;
		
	read_seis_dat();
	seismic->redraw();
}

void doNext(Fl_Widget*, void*)
{
	if(tr_fin<Ntr)
	{
		tr_fin += tr_vis;			// tr_vis remains unchanged
		if(tr_fin>Ntr) tr_fin=Ntr;
		tr_sta = tr_fin-tr_vis+1;	// from tr_vis formula

		sprintf(cgbuf,"%d",tr_sta);  tr_sta_field->value(cgbuf);
		sprintf(cgbuf,"%d",tr_fin);  tr_fin_field->value(cgbuf);
	
		read_seis_dat();
		seismic->redraw();
	}
}

void do_wiggle(Fl_Widget*, void*)
{
	if(wiggle_trace==0) wiggle_trace=1;
	else wiggle_trace=0; 
	seismic->redraw();
}

////////////////////////////////////////////////////////////////////////
/////////////////    Window with PROCESSED data     ////////////////////	
////////////////////////////////////////////////////////////////////////
void Process(Fl_Widget *, void*)
{  

	if(nmo_no>3 || nmo_no<1)
	{			
		fl_message("WARNING: Undefined NMO correction method [# %d]\n\
Close the window to continue...",nmo_no);
		return;
	}	
	
	switch(nmo_no)
	{
		case 1: strcpy(cgbuf,"Hyperbolic reflection moveout applied"); 
				break;
		case 2: strcpy(cgbuf,"Non-hyperbolic moveout in anisotropic media"); 
				break;
		case 3: strcpy(cgbuf,"Non-Hyperbolic moveout in isotropic media"); 
				break;
		default: strcpy(cgbuf,"Default reflection moveout applied");
				break;
	}		

	proc_window = new Fl_Double_Window(canvas_w+RIM*2,canvas_h+RIM*2,cgbuf);
			
	proc_draw = new ProcCanvas(RIM,RIM,canvas_w,canvas_h);
	//seismic->redraw();	

	proc_window->end();
	proc_window->resizable(proc_window); // Critical !!!
	proc_window->show();
}

/////////////////////////////////////////////////////////////////////
/****   MAIN   MAIN   MAIN   MAIN   MAIN   MAIN   MAIN   MAIN   ****/
/////////////////////////////////////////////////////////////////////
int main(int argc, char** argv)
{
	int i,tr;
	double t,val;
	char cbuf1[90],cbuf2[90],cbuf3[90],cbuf4[90];
	char cbuf[90],plot_hdr[120];	
	///////////////////////////////////////////////////////////
	
	Fl::scheme(NULL); //  "plastic" for Aqua and "gtk+"
	// NULL use the scheme defined in the FLTK_SCHEME env.var. 
	
	if(argc == 2) strcpy(inpname,argv[1]);		// Do not use terminal
	else		// if you want to create an Apple  *.app bundle application
	{
		File_menu *choice = new File_menu();
		choice->show();
		choice->end(); // no more widgets will be added to the window
		while(tr_vis<0) Fl::wait();
		choice->hide();
	}

	read_file_hdr(inpname);
	read_seis_dat();			// into new seis[]

	///////////////////////////////////////////////////////////
	strcpy(plot_hdr,"Original Data");
	seis_window = new Fl_Double_Window(canvas_w+RIM*2,canvas_h+RIM*2,plot_hdr);
	
	seismic = new Draw_seis(RIM,RIM,canvas_w,canvas_h);
	// seismic->end();	
	
	seis_window->end();
	seis_window->resizable(seis_window); // Critical !!!
	seis_window->show();

	///////////////////////////////////////////////////////////
		
	opts_window = new Fl_Window(1050,50,420,972,"Display Manager");
	// The follwing widgets belong to this window	

	Fl_Box* label_o = new Fl_Box(FL_FRAME_BOX,20,10,380,24,
					"==========   Trace Display Parameters   ==========");
	label_o->labelfont(0);  label_o->labelcolor(7);
	label_o->color(36);		label_o->labelsize(14);
	//label_o->labeltype(FL_ENGRAVED_LABEL);

	Fl_Box* label_tr = new Fl_Box(80,30,140,50,"Range of displayed traces :");
	label_tr->labelfont(FL_HELVETICA_BOLD);
	label_tr->labelsize(14);
	
	tr_sta_field = new Fl_Int_Input(80, 70, 70, 25, " from ");
	sprintf(cbuf,"%d",tr_sta);  tr_sta_field->value(cbuf);
	tr_sta_field->callback(input_tr_sta);
	tr_fin_field = new Fl_Int_Input(180, 70, 70, 25, " to ");
	sprintf(cbuf,"%d",tr_fin);  tr_fin_field->value(cbuf);
	tr_fin_field->callback(input_tr_fin);
	
	next_tr = new Fl_Button(285, 70, 100, 25, " Next traces");
	next_tr->color(62);
	next_tr->labelfont(1);	
	next_tr->callback(doNext);

	sprintf(cbuf1,"Total number of traces = %d",Ntr);	
	Fl_Box* comment1 = new Fl_Box(80,82,150,50,cbuf1);
	comment1->labelfont(FL_HELVETICA);
	comment1->labelsize(13);		
		
	Fl_Box* label_t = new Fl_Box(80,110,140,50,"Time range [ms] to display:");
	label_t->labelfont(FL_HELVETICA_BOLD);
	label_t->labelsize(14);	
	
	t_sta_field = new Fl_Int_Input(80, 150, 70, 25, " from ");
	sprintf(cbuf,"%d",samp_sta*dsamp/1000); t_sta_field->value(cbuf); 
	t_sta_field->callback(input_t_sta);
	t_fin_field = new Fl_Int_Input(180, 150, 70, 25, " to ");
	sprintf(cbuf,"%d",samp_fin*dsamp/1000); t_fin_field->value(cbuf); 
	t_fin_field->callback(input_t_fin);
	
	gain_field = new Fl_Float_Input(310, 150, 70, 25, "Gain: ");
	sprintf(cbuf,"%5.2f",gain);  gain_field->value(cbuf);
	gain_field->labelsize(14);
	gain_field->labelfont(FL_HELVETICA_BOLD);	
	gain_field->callback(input_gain);
	
	sprintf(cbuf4,"Total trace length = %d msec",nsamp*dsamp/1000);	
	Fl_Box* comment4 = new Fl_Box(80,160,150,50,cbuf4);
	comment4->labelfont(FL_HELVETICA);
	comment4->labelsize(13);	

	show_hdr = new Fl_Button(200,200,180,25,"Show file's text header");
	show_hdr->color(62);									
	show_hdr->labelcolor(FL_BLACK);
	show_hdr->labelfont(3);	
	show_hdr->labelsize(14);	
	show_hdr->callback(show_hdr_cb);	
	
	Fl_Box* label_hdr = new Fl_Box(50,225,300,54,
									"Select header plot parameter\n\
from the list of SEGY trace headers:");
	label_hdr->labelfont(FL_HELVETICA_BOLD);
	label_hdr->labelsize(14);
	Fl_Input_Choice in_hdr(50,270,300,30,"");
	in_hdr.callback(input_choice_cb, (void*)&in_hdr);
	
	in_hdr.add("  2 - Field record number");
	in_hdr.add("  3 - Trace number within Field record");
	in_hdr.add("  4 - Energy source point number"); 
	in_hdr.add("  5 - Ensemble number (CDP, CRT, ...)");
	in_hdr.add("  6 - Trace number within the ensemble");
	in_hdr.add("  9 - Source-receiver offset");
	in_hdr.add(" 10 - Receiver group elevation");  
	in_hdr.add(" 11 - Surface elevation at source");
	in_hdr.add(" 13 - Datum elevation at receiver");
	in_hdr.add(" 14 - Datum elevation at source");
	in_hdr.add(" 18 - Source coordinate X");
	in_hdr.add(" 19 - Source coordinate Y");
	in_hdr.add(" 20 - Group coordinate X");
	in_hdr.add(" 21 - Group coordinate Y");
	in_hdr.add(" 45 - ensemble (CMP) coordinate X");
	in_hdr.add(" 46 - ensemble (CMP) coordinate Y");
	in_hdr.add(" 47 - In-line number for 3-D data");
	in_hdr.add(" 48 - Cross-line number for 3-D data");
	in_hdr.add(" 49 - Shotpoint number (for 2-D data)");		
	in_hdr.value(3);  // CDP - default header plot parameter
	
	Fl_Box* comment = new Fl_Box(50,292,300,50,
				"the leading number indicates 4-byte integer word,\n\
starting from 0, within the 240-byte trace header");
	comment->labelfont(FL_HELVETICA);
	comment->labelsize(13);
	
	Fl_Spinner* spin = new Fl_Spinner(250, 345, 70, 30, 
										"Header spacing, traces ");
	spin->labelsize(14);
	spin->labelfont(FL_HELVETICA_BOLD);	
	spin->minimum(1);
	spin->maximum(250);
	spin->step(5);
	spin->value(hdr_step);	
	spin->align(Fl_Align(FL_ALIGN_LEFT));
	spin->callback(spinner_cb);
	
	apply = new Fl_Button(60,390,300,30,
								"Redraw traces with new parameters");
	apply->color(62);
	apply->labelcolor(FL_BLACK);
	apply->labelfont(3);	
	apply->labelsize(16);	
	apply->callback(doRedraw);	
	
	
	add_wiggle = new Fl_Button(160, 430, 200, 25, 
										"Add/remove wiggle trace");
	add_wiggle->color(62);									
	add_wiggle->labelcolor(FL_BLACK);
	add_wiggle->labelfont(3);	
	add_wiggle->labelsize(14);										
	add_wiggle->callback(do_wiggle);
	
	////////////////////////////////////////////////////////////////////////
	////////////////////      Normal Moveout        ////////////////////////	
	////////////////////////////////////////////////////////////////////////
		
	Fl_Box* label_oo = new Fl_Box(FL_FRAME_BOX,20,470,380,24,
					"=========    Normal Moveout Parameters    =========");
	label_oo->labelfont(0);  label_oo->labelcolor(7);
	label_oo->color(36);		label_oo->labelsize(14);
	
	Fl_Box* nmo_param = new Fl_Box(80,500,260,50,
							"Specify parameters for NMO application\n\
considering a power-law velocity function\n\
V = v0 + alpha * t^beta");
	nmo_param->labelfont(FL_HELVETICA_BOLD);
	nmo_param->labelsize(14);		
				
	
	v0_field = new Fl_Float_Input(100, 565, 50, 25, " V0 = ");
	sprintf(cbuf,"%.0f",v0); v0_field->value(cbuf);
	v0_field->labelsize(15); v0_field->labelfont(FL_HELVETICA_BOLD);
	v0_field->callback(input_v0);
	Fl_Box* comment2 = new Fl_Box(170,553,200,50,
						"Velocity at zero time [ m/s or ft/s ]");
	comment2->labelfont(FL_HELVETICA);  comment2->labelsize(14);	
	
	
	alfa_field = new Fl_Float_Input(100, 600, 50, 25, " alpha = ");
	sprintf(cbuf,"%.0f",alfa); alfa_field->value(cbuf);
	alfa_field->labelsize(15); alfa_field->labelfont(FL_HELVETICA_BOLD);	
	alfa_field->callback(input_alfa);
	Fl_Box* comment5 = new Fl_Box(170,588,200,50,
							"Velocity gradient [ m/s or ft/s ]");
	comment5->labelfont(FL_HELVETICA);  comment5->labelsize(14);	
	
		
	beta_field = new Fl_Float_Input(100, 635, 50, 25, " beta = ");
	sprintf(cbuf,"%.2f",beta); beta_field->value(cbuf);
	beta_field->labelsize(15); beta_field->labelfont(FL_HELVETICA_BOLD);	
	beta_field->callback(input_beta);
	Fl_Box* comment6 = new Fl_Box(170,623,200,50,
							"Velocity curvature [ dimensionless ]");
	comment6->labelfont(FL_HELVETICA);  comment6->labelsize(14);		


	maxs_field = new Fl_Float_Input(250, 665, 50, 25,
										"Stretch Factor [ % ] = ");
	sprintf(cbuf,"%3.0f",maxSF);  maxs_field->value(cbuf);
	maxs_field->labelsize(15);  maxs_field->labelfont(FL_HELVETICA_BOLD);	
	maxs_field->callback(input_maxs);	
	
	Fl_Box* comment3 = new Fl_Box(50,682,320,50,
		"Is used to suppress normal moveout stretch artifacts,\n\
commonly set between 15 and 20 percent.");
	comment3->labelfont(FL_HELVETICA);  comment3->labelsize(14);
	
	
	nu_field = new Fl_Float_Input(280, 730, 50, 25,
								"Anisotropy parameter [ % ] = ");
	sprintf(cbuf,"%3.0f",nu);  nu_field->value(cbuf);
	nu_field->labelsize(15);  nu_field->labelfont(FL_HELVETICA_BOLD);	
	nu_field->callback(input_nu);	
	
	Fl_Box* comment7 = new Fl_Box(50,747,320,50,
	"Utilizes non-hyperbolic reflection traveltime approximation\n\
for transversely isotropic media given by Alkhalifah [1997].");
	comment7->labelfont(FL_HELVETICA);  comment7->labelsize(14);	
	
	
	Fl_Box* NMO_type = new Fl_Box(50,780,300,54,
									"Select NMO correction method:");
	NMO_type->labelsize(15);  NMO_type->labelfont(FL_HELVETICA_BOLD);
	Fl_Input_Choice in_nmo(45,820,330,30,"");
	in_nmo.callback(nmo_choice_cb, (void*)&in_nmo);
	
	in_nmo.add(" 1 - hyperbolic normal moveout");
	in_nmo.add(" 2 - non-hyperbolic, transversely isotropic media");	
	in_nmo.add(" 3 - non-hyperbolic, isotropic layered media ");
	in_nmo.value(0);  // hyperbolic - default correction method	
	
	//////////   Callback function to open Processed window    /////////////
	apply_nmo = new Fl_Button(90, 870, 240, 30, 
									 " Apply NMO corrections ");
	apply_nmo->color(62); apply_nmo->labelsize(16); apply_nmo->labelfont(3);
	apply_nmo->labelcolor(FL_BLACK);          apply_nmo->callback(Process);	
	
	////////////////////////////////////////////////////////////////////////	
	////////////////////////////////////////////////////////////////////////		
	end_this = new Fl_Button(165,920, 80, 30, "Exit");
	end_this->color(88);
	end_this->labelcolor(FL_BLACK);
	end_this->labelfont(3);	
	end_this->labelsize(18);	
	end_this->callback(exit_b);	
			
	opts_window->end(); 
	opts_window->show();

	return(Fl::run());
	
	fclose(in_fp);
   return 0;
}
/////////////////////////////////////////////////////////////////////
/****    END_MAIN   END_MAIN   ENF_MAIN   END_MAIN   END_MAIN   ****/
/////////////////////////////////////////////////////////////////////
/****               Called Function Definition                  ****/
/////////////////////////////////////////////////////////////////////
void read_file_hdr(const char *name)
{
	int i=1,j,tr;
	long file_size,offset;
	float amp,count,avr;	
	char array[3601];
	short *hdr2,format,measure;	
	char formats[][30] =    // string formats[9]
	{
	   "",
	   "4-byte IBM floating-point",     // 1
	   "4-byte integer",
	   "2-byte integer", 
	   "4-byte fixed-point with gain",
	   "4-byte IEEE floating-point",    // 5
	   "byte-swapped IEEE f.p.",        // 6
	   "non-standard 4-byte IEEE f.p.",  
	   "1-byte integer"
	};
	
	if( *( (char *)&i ) != 1 )	// TEST on COMPUTER ENDIANISM
	{
		fl_message("This application is supposed to run on\n\
Intel-based PC's with little-Endian data format.\n\
This is big-Endian computer. Close the window to exit");
		exit(1);
	}
	
	if ((in_fp = fopen(name, "rb")) == NULL)
	{fl_message("ERROR opening input file!\nClose the window to exit");exit(1);}
	
	fseek(in_fp, 0L, SEEK_END);	// go to file's end
	file_size = ftell(in_fp);
	fseek(in_fp, 0L, SEEK_SET);	// go to beginning	
	
   fread (&array, 3600, 1, in_fp);	// 1 object 3600 bytes long
	
	//   Convert Textual File Header:  EBCDIC-to-ASCII
	for(i=0;i<3200;i++) text_hdr[i]=to_ascii(array[i]);
	for(i=1;i<3200;i++) if(i%80==0) text_hdr[i-1]='\n';
	
   hdr2=(short*)array;
   format=hdr2[1612];			// Correspond to byte 3225
   
   if(format>0 && format<9) ; 	// Native format - do nothing
   else							// swap bytes (for both IBM and IEEE)
   {
		for(i=0;i<24;i++) convert2(array+3212+i*2);
      format=hdr2[1612];		
		if(format>0 && format<9) ;
		else 
		{
			fl_message("ERROR: input is not a SEG-Y file!\nClose the window to exit");
			exit(1);
		}
   }   
	
   dsamp=(int)hdr2[1608];	// SAMPLE_INTERVAL at byte 3217 <-- GLOBAL
   nsamp=(int)hdr2[1610];	// NUMBER_SAMPLES at byte 3221  <-- GLOBAL
   measure=hdr2[1627];		// Measurement system at byte 3255
	
   if(dsamp<0 || dsamp>8000 || nsamp<10 || nsamp>16000)
   { 
		fl_message("ERROR:  irrational sampling information.\n\
Sample Interval = %d microseconds\nNumber of Trace Samples = %d\n\
Close the window to exit...",dsamp,nsamp);	
		fclose(in_fp);   exit(1);
   }
	
	tr_size = 60*sizeof(int) + nsamp*sizeof(float);
	
	Ntr = (int)((file_size-3600L)/(nsamp*sizeof(float)+240));
	
   convert_mode=1;  				// No conversion
   if(format==5) convert_mode=2; 	// IEEE-to-native
   if(format==1) convert_mode=3;	// IBM--to-native
	
	///////////////////////////////////////////////////////////
	/****       Initial Visualization Parameters          ****/
	
	tr_sta=1;
	if(Ntr<50) tr_fin=4;
	else tr_fin=50;
	
	tr_vis = tr_fin-tr_sta+1; // if trace count starts from '1'
	samp_sta=0;  samp_fin=10;
	if(nsamp<100) samp_fin=10;
	else if(nsamp<500) samp_fin=100;
		else if(nsamp<1002) samp_fin=500;
			  else samp_fin=1000;				
	samp_vis = samp_fin-samp_sta;
	gain = 1.0;
	
	/////////////////////////////////////////////////////////////
	/**      Get Normalization Factor for visualization       **/
	/**   by reading data until non-zero trace will be found  **/
	
	hdrs = new (nothrow) char[240];	
	seis = new (nothrow) float[nsamp];
	
	Norm=1.0;
	for(tr=0;tr<Ntr;tr++)
	{
		offset=3600L+tr_size*tr;	
		fseek(in_fp, offset, SEEK_SET);
		fread (hdrs, 240, 1, in_fp);
		if(convert_mode>1) convert_hdrs(hdrs);
		hdr2=(short*)hdrs; 
		if(hdr2[14]==2) continue;	// Trace ID code (0=Unknown,1=Seismic,2=Dead)	
		fread (seis, nsamp*sizeof(float), 1, in_fp);
		if(convert_mode>1)
		{
			if(convert_mode==2)			// IEEE-to-native
				for(i=0;i<nsamp;i++) seis[i]=efloat(seis[i]);
			if(convert_mode==3)			// IBM-to-native
				for(i=0;i<nsamp;i++) seis[i]=ibm2ieee(seis[i]);			
		} 	
		
		count=avr=0.0;
		for(i=0;i<nsamp;i++) 
		{ 
			amp=fabs(seis[i]); 
			if(amp>0.0) { avr+=amp; count+=1.0; }
		}
		if(count>0.0) { avr/=count; Norm=4.0*avr; break; } 
	}
	
	if(hdr2[34] >= 0) elev_scal=(float)hdr2[34];
	else elev_scal=(float)(-1.0/hdr2[34]);
	if(hdr2[35] >= 0) coor_scal=(float)hdr2[35];
	else coor_scal=(float)(-1.0/hdr2[35]);	
	
	//////////////////////////////////////////////////////////////////
	//////////////   Info Screen Message   /////////////////////////// 
   if(measure<1 || measure>2) strcpy(cgbuf,"unknown");
   if(measure==1) strcpy(cgbuf,"metric, meters");
   if(measure==2) strcpy(cgbuf,"imperial, feet");	
	
	fl_message("File \"%s\"\nFile size = %ld kilobytes\nFile Header Info:\n\
======================================\n\
SEGY format code = %d ==> %s\n\
Sample Interval = %d microseconds\nNumber of Trace Samples = %d\n\
Measurement system = %d ==> %s\nNumber of traces = %d\n\
Average trace magnitude = %g\nElevation scalar value = %g\n\
Coordinate scalar value = %g\n\
======================================\n\
Close the window to continue...",
name,file_size/1000,format,formats[format],dsamp,nsamp,measure,\
cgbuf,Ntr,avr,elev_scal,coor_scal);
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////	
}
////////////////////////////////////////////////////////////////
void read_seis_dat()		// fill up seis[] with data; float *seis;
{							// Do memory re-allocation before reading
	int i,tr;
	long offset;	
	
	delete[] hdrs; // originally allocated in read_file_hdr();	
	hdrs = new (nothrow) char[tr_vis * 240];		
	
	delete[] seis; // originally allocated in read_file_hdr();
	seis = new (nothrow) float[(tr_vis+1) * nsamp];
	if(seis == 0)
	{ 
		fl_message("ERROR: NO memory to accomodate seismic data!\n\
Close the window to exit...");	
		fclose(in_fp);  exit(1);
	}
	
/****   seis[..] contains the whole trace from t=0 to 'nsamp'   ****/

	for(tr=0;tr<tr_vis;tr++)
	{	// tr_size = 60*sizeof(int) + nsamp*sizeof(float);
		offset = 3600L + tr_size*(tr_sta+tr-1);	
		fseek(in_fp, offset, SEEK_SET);
		fread (hdrs+tr*240, 240, 1, in_fp); // 240 bytes-long object
      if(convert_mode>1) convert_hdrs(hdrs+tr*240);		
		// fseek(in_fp, samp_sta*sizeof(float), SEEK_CUR);		
		fread (seis+tr*nsamp, nsamp*sizeof(float), 1, in_fp);
		if(convert_mode>1)
		{
			if(convert_mode==2)			// IEEE-to-native
				for(i=0;i<nsamp;i++) 
					seis[tr*nsamp+i]=efloat(seis[tr*nsamp+i]);
			if(convert_mode==3)			// IBM-to-native
				for(i=0;i<nsamp;i++) 
					seis[tr*nsamp+i]=ibm2ieee(seis[tr*nsamp+i]);			
		} 		
	}	
}
////////////////////////////////////////////////////////////////////
//	 Resampling using Cubic Convolution Method [R. Keys, 1981].
// Four data values from an array  dat[0..NS1]  are used every
// time to RETURN a calculated "index" value of the RESAMPLED
// array having size NS2 and same origin as the input array.
////////////////////////////////////////////////////////////////
float conv3(float *dat, int NS1, int NS2, int index)
{
	float y_1,y0,y1,y2,dx,dx2,dx3,k_1,k0,k1,k2;
	
	if(index<0 || index >=NS2) return 0.0;
	
	int i = index*NS1/NS2;
	
	y0=dat[i]; 	
	if(i<1) y_1=y0; else y_1=dat[i-1];
	if(i>NS1-2) y1=y0; else y1=dat[i+1]; 
	if(i>NS1-3) y2=y1; else y2=dat[i+2];
	
	dx = (float)(index*NS1 - i*NS2)/NS2;		
	dx2=dx*dx;  dx3=dx*dx2;		
	
	k_1 = (2.0*dx2 - dx3 - dx)/2.0;
	k0 = (3.0*dx3 - 5.0*dx2 + 2.0)/2.0;
	k1 = (4.0*dx2 - 3.0*dx3 + dx)/2.0;
	k2 = (dx3 - dx2)/2.0;
	return (k_1*y_1 + k0*y0 + k1*y1 + k2*y2);	
}
//
void convert_hdrs(char* array)  //  Byte-swapping for the
{                               //  240-byte array of trace headers
   int k;
   void convert2(char* input);
   void convert4(char* input);
	
   for(k=0;k<7;k++) convert4(array+k*4);       // Bytes   1 - 28
   for(k=0;k<4;k++) convert2(array+28+k*2);    // Bytes  29 - 36
   for(k=0;k<8;k++) convert4(array+36+k*4);    // Bytes  37 - 68
   for(k=0;k<2;k++) convert2(array+68+k*2);    // Bytes  69 - 72
   for(k=0;k<4;k++) convert4(array+72+k*4);    // Bytes  73 - 88
   for(k=0;k<46;k++) convert2(array+88+k*2);   // Bytes  89 - 180
   for(k=0;k<5;k++) convert4(array+180+k*4);   // Bytes 181 - 200
   for(k=0;k<2;k++) convert2(array+200+k*2);   // Bytes 201 - 204
   convert4(array+204);  convert2(array+208);  // Bytes 205 - 210
   for(k=0;k<5;k++) convert2(array+210+k*2);   // Bytes 211 - 224
   convert4(array+224);  convert2(array+228);  // Bytes 225 - 230
   convert2(array+230);                        // Bytes 231 - 233
	// Unassigned bytes 233 - 240
}
////////////////////////////////////////////////////////////////
//
// SUBROUTINES
// WRITTEN BY MARK GOLDMAN
// U.S. GEOLOGICAL SURVEY
// http://pubs.usgs.gov/of/2005/1311/
//
// Adopted by Yuri Ganshin
//
void convert2(char* input) {
	
   unsigned char temp;
	
   temp = input[1];
   input[1] = input[0];
   input[0] = temp;
}

void convert4(char* input) {
	
   unsigned char temp;
	
   temp = input[3];
   input[3] = input[0];
   input[0] = temp;
	
   temp = input[2];
   input[2] = input[1];
   input[1] = temp;
}

float efloat(float value) {
	
   union {
      char bytes[4];
      float flt;
   } ieee;
	
	// swap bytes
   unsigned char temp;
   ieee.flt = value;
	
   temp = ieee.bytes[3];
   ieee.bytes[3] = ieee.bytes[0];
   ieee.bytes[0] = temp;
	
   temp = ieee.bytes[2];
   ieee.bytes[2] = ieee.bytes[1];
   ieee.bytes[1] = temp;

   return(ieee.flt);
}

float ibm2ieee(const float ibm) 
{
   long IntMantissa;
   float Mantissa,PosResult;
   // DEFINE AND SETUP UNION VARIABLES
   union { float f; unsigned char c[4]; } src; 
	
   // ASSIGN INPUT PARAMETER TO UNION MEMBER
   src.f = ibm; // When you change the value of a union member, 
	// you modify all the members of the union
   // CONVERTION
   IntMantissa = ((long)src.c[1] << 16)+((long)src.c[2]<<8)+src.c[3];
   Mantissa = (float)(IntMantissa)/(float)(0x1000000);
   PosResult = Mantissa*pow(16.0,(double)((src.c[0]&0x7F)-64));
	
   if (src.c[0] & 0x80) return -PosResult;
   else return PosResult;
}

/*
 **  ASCII <=> EBCDIC conversion functions
 */
unsigned char to_ascii ( unsigned char c )
{
	unsigned char e2a[256] = {
		0,  1,  2,  3,156,  9,134,127,151,141,142, 11, 12, 13, 14, 15,
		16, 17, 18, 19,157,133,  8,135, 24, 25,146,143, 28, 29, 30, 31,
		128,129,130,131,132, 10, 23, 27,136,137,138,139,140,  5,  6,  7,
		144,145, 22,147,148,149,150,  4,152,153,154,155, 20, 21,158, 26,
		32,160,161,162,163,164,165,166,167,168, 91, 46, 60, 40, 43, 33,
		38,169,170,171,172,173,174,175,176,177, 93, 36, 42, 41, 59, 94,
		45, 47,178,179,180,181,182,183,184,185,124, 44, 37, 95, 62, 63,
		186,187,188,189,190,191,192,193,194, 96, 58, 35, 64, 39, 61, 34,
		195, 97, 98, 99,100,101,102,103,104,105,196,197,198,199,200,201,
		202,106,107,108,109,110,111,112,113,114,203,204,205,206,207,208,
		209,126,115,116,117,118,119,120,121,122,210,211,212,213,214,215,
		216,217,218,219,220,221,222,223,224,225,226,227,228,229,230,231,
		123, 65, 66, 67, 68, 69, 70, 71, 72, 73,232,233,234,235,236,237,
		125, 74, 75, 76, 77, 78, 79, 80, 81, 82,238,239,240,241,242,243,
		92,159, 83, 84, 85, 86, 87, 88, 89, 90,244,245,246,247,248,249,
		48, 49, 50, 51, 52, 53, 54, 55, 56, 57,250,251,252,253,254,255
	};
	
	return e2a[c];
}
